// --- [ DCF_77.c ] -------------------------------------------------------------
//       26.03.10 - reduzierte Version
//       tab = 3
//
// ----------------------------------------------------------------------------
#define EXTERN

#include <avr/io.h>
#include <inttypes.h>
#include <stdlib.h>
#include <string.h>
#include <util/delay.h>
#include <avr/eeprom.h>
#include <avr/interrupt.h>
#include <avr/pgmspace.h>
#include <avr/wdt.h>

#include "global.h"
#include "dcf_bits.h"
#include "dcf_77.h"
#include "lcd.h"
#include "rs232.h"

uint8_t summand[] 		PROGMEM = { 1, 2, 4, 8, 10, 20, 40, 80 }; // fr dcf77
uint8_t days_of_month[] PROGMEM = { 0, 31, 28, 31, 30, 31, 30, 31, 31, 30, 31, 30, 31 };
char day_names[] 			PROGMEM = {'M','o','D','i','M','i','D','o','F','r','S','a','S','o' };

volatile uint8_t flag;						// sammelt Status-Flags
volatile uint8_t error;						// sammelt error-Flags
volatile uint8_t T0_ticks;					// Counter fr die interne Uhr / Sekundensignal
volatile uint8_t T1_ticks;					// Counter fr die Lnge des DCF-Signals (in 0.01sek)
uint8_t error_abs;							// summiert die Anzahl der fehlerhaften Minuten
													// wird um 0.00h zurckgesetzt

// ----------------------------------------------------------------------------
int main(void)
{
// --- INITIALISIERUNG --------------------------------------------------------
uint8_t mcusr;									// Flags in MCUSR nach dem Start
uint8_t DCF_valid = TRUE;
uint8_t synch_time = 0;						// nach dem Programmstart eine Synchronisation erzwungen wird

// --- PORTs ------------------------------------------------------------------
// PD.0/1	RxD, TxD 						// fr serielle Schnittstelle
// PD.2/3	INT0, INT1						// hier wird die Uhr angeschlossen
// PD.4     frei (1)
// PB.6/7	XTAL								// der Quarz bentigt zwei Pin
// PD.5..7	frei (3)							// PD.7 als Schalter genutzt -> DAYLIGHT_PIN
// PB.0..2  LCD 								// Steuerleitungen LCD siehe lcd.h
// PB.4/5   frei (2)
// PC.0..3  LCD								// Datenleitungen LCD  siehe lcd.h
// PC.4/5	SCL,SDA							// TWI-Port

// erwartet wird vom DCF-Empfnger:
// DCF-Ruhe-Level: HIGH, bei Signal: LOW !!!!!!!!!!!!!!!!!!!!!!!!!

mcusr = MCUSR;									// MCUSR sichern
MCUSR = 0x0F;									// alle Flags in MCUSR lschen


//DDRD  = 1<<DAYLIGHT_PIN;					// nur PIN.7 als Alarmausgang benutzt
//DCF_set_daylight_pin(1);					// den Pin auf HIgh
													// PD.0/PD.1 = RxD/TxD
//DDRD = (1<<PD0| 1<<PD1 | 1<<PD4 | 1<<PD5| 1<<PD6| 1<<PD7);

// --- ADC --------------------------------------------------------------------
PRR = (1<<PRSPI | 0<<PRUSART0 | 1<<PRADC);// SPI, (USART) und ADC abschalten

// --- EXT_INT ----------------------------------------------------------------
#ifdef M168
EICRA = (1<<ISC11 | 1<< ISC10 | 1<< ISC01);// Rising INT1, Falling INT0
EIMSK  = (1<<INT1 | 1<< INT0);			// INT0 / INT1_isr enablen
#endif
#ifdef M8
MCUCR = (1<<ISC11 | 1<< ISC10 | 1<< ISC01);// Rising INT1, Falling INT0
GICR  = (1<<INT1 | 1<< INT0);			// INT0 / INT1_isr enablen
#endif

// --- CLock-Prescaler --------------------------------------------------------
#ifdef M168
CLKPR = (1<<CLKPCE);
CLKPR &= 0xF0;									// lower nibble lschen -> Faktor 1
#endif

// --- TIMER0 -----------------------------------------------------------------
// generiert den internen 1-Sekunden-Takt
// 3.686.400 / ((64 * 256) * 225) = 1 Sekunde
TCCR0B = (1<<CS01 | 1<< CS00);			// Prescale 64, den Rest macht TIMER0_ISR
#define T0_PRELOAD	225

// --- TIMER1 -----------------------------------------------------------------
// luft als Zhler fr Pulsdauer und Pulsabstand
// 3.686.400 / (256 * 144)`= 0.01 Sekunde
TCCR1A = 0;
TCCR1B = (1<<CS12);							// Prescale 256
OCR1AH = 0;
OCR1AL = 144;

// --- beide Timer ------------------------------------------------------------
#ifdef M8
//TIMSK = (1<<TOIE0 | 1<<OCIE1A);		// ISR fr Timer1 / 2 enablen
#endif
#ifdef M168
TIMSK0 = (1<<TOIE0 );						// ISR fr Timer1 enablen
TIMSK1 = (1<<OCIE1A);						// ISR fr Timer2 enablen
#endif

asm volatile("sei");

#ifdef LOG_RS232
rs232_init(TEILER_9600);
put_c(13);
puts_p(PSTR("Logging on RS232"));		// "Hallo World" ausgeben
put_c(13);

#endif

// --- intro ------------------------------------------------------------------
// liefert im gnstigsten Fall nach 62 Sekunden ein vollstndiges Datum,
// im ungnstigsten Fall nach 120 Sekunden


#ifdef LCD
lcd_init(LCD_DISP_ON);
lcd_puts_p(PSTR("wait for synch .."));// Hinweis ausgeben

lcd_goto_xy(2,1);
lcd_puts_p(PSTR("____________________"));// Linie ausgeben


#ifdef WATCHDOG
if (mcusr & (1<<WDRF))						// wenn sich ein watchdog-reset ereignet hat
	{
	MCUSR = 0x0F;								// Flag lschen
#ifdef LCD
	lcd_goto_xy(3,1);
	lcd_puts("wd");							// "wd" in Zeile 3 ausgeben
#endif
#ifdef LOG_RS232
	put_c(13);
	puts_p(PSTR("WD-reset"));				// Meldung ber WDR ausgeben
	put_c(13);
#endif

	}
#endif

#endif

// --- warten auf den Minutenanfang -------------------------------------------
//
while (!(flag & (1<<DCF_MINUTE)))		// solange warten, bis das Minutenflag gesetzt ist
	{
	if (flag & (1<<XTL_SEC))				// die Sekunden seit dem Start anzeigen
		{
		flag &= (~(1<<XTL_SEC));
		if (error) error = 0;				// Fehler sollen die Synchronisation mit dem
													// Minutenbeginn nicht stren
													// Timer0 setzt bei Fehler kein Minutenflag
		#ifdef LCD
		display_second();						// nur die Sekundenanzeige aktualisieren
													// Damit der Zhler weiterluft, wenn bis hierhin
													// kein Minutensignal gefunden wurde.
		if (time[CLK][SEKUNDE] == 58) time[CLK][SEKUNDE] = 0;
		#endif
		}
	}

// --- Minutenanfang erkannt --------------------------------------------------
//
#ifdef ZEIT_ZEICHEN
time[CLK][STUNDE] = 16;						// damit das Zeitzeichen fr 4:00 ausgegeben wird
zeit_zeichen();								// gibt das Zeitzeichen aus
#endif
													// jetzt ist das Minutenflag der DCF-Uhr gesetzt
time[CLK][STUNDE]  = 0;
time[CLK][SEKUNDE] = 0;						// wieder bei Sekunde 0 starten,
DCF_reset_bits();								// den Index von DCF_bits auf 0 setzen
error 	= 0;
													// Flag fr Neustart und Neusychronisation setzen
flag 		= ( 1<<START_FLAG | 1<<SYNCH_ENABLE );
T1_ticks = 0;									// alle Counter zurcksetzen
TCNT1H   = 0;
TCNT1L   = 0;

#ifdef LCD
lcd_goto_xy(1,1);
lcd_puts_p(PSTR("new minute found:"));	// Hinweis ausgeben
lcd_goto_xy(4,1);
lcd_puts_p(PSTR("            "));		// Text 'TWI-Adr' lschen
#endif


// --- Watchdog ---------------------------------------------------------------
// den Watchdog erst dann enablen, wenn ein DCF-Signal gefunden ist
// sonst wird bei fehlendem Signal nach 59 Sekunden ein reset ausgefhrt.
// der Watchdog wird zurckgesetzt in main bei flag &= (1<<XTL_SEC)
#ifdef WATCHDOG
asm volatile("cli");
WDTCSR = (1<<WDCE | 1<<WDE);
WDTCSR = (0<<WDCE | 1<<WDE | 1<<WDP3);	// 4 Sekunden (mindestens 2 Sekunden sind erforderlich
asm volatile("sei");							// um die Zeit von der 59. Sekunde bis zur 1.Sekunde zu berbrcken)
#endif

// --- main -------------------------------------------------------------------
// eine Endlosschleife, in der die Flag.bits den Ablauf steuern

while(1)
	{
	if (error) DCF_valid = FALSE;			// sobald ein Fehler aufgetreten ist,
													// wird auf die xtal-Uhr umgeschaltet.

// --- neue Sekunde der xtal-Uhr ----------------------------------------------

	if (flag & (1<<XTL_SEC))				// das Flag fr eine neue Sekunde ist gesetzt
		{
		flag &= (~(1<<XTL_SEC));
		#ifdef LCD
		display_second();						// nur die Sekundenanzeige aktualisieren
		#endif
		#ifdef WATCHDOG
		asm volatile("wdr");
		#endif
													// Bei einem Minutenwechsel werden zwei Flags gesetzt,
													// XTL_MINUTE und DCF_MINUTE, die Reihenfolge ist zufllig.
													// Um zu verhindern, dass nach dem ndern von SYNCH_ENABLE
													// beide zuschlagen und die Zeit 2x incrementiert wird, wird
		if (time[CLK][SEKUNDE] == 55)		// das Flag erst so spt wie mglich gendert.
			{								
			if (synch_time)
				{									// Das Zeitintervall bis zum nchsten Synchronisationsversuch
				synch_time--;					// decrementieren - bei 0 wird synchronisiert
				flag &= (~(1<<SYNCH_ENABLE));	// das Flag zur Synchronisation lschen
				}
			else
				{
				flag |= (1<<SYNCH_ENABLE);	// bei 0 das Flag setzen
													// wenn das Flag DCF_MINUTE nach dem Flag XTL_MINUTE gesetzt wird,
													// kann es einen zustzlichen Minutenanfang auslsen, da jetzt SYNCH_ENABLE
													// gesetzt ist.
													// Um das zu verhindern, wird das Flag prophylaktisch gelscht und 
													// DCF_reset_bits() aufgerufen (was normalerweise durch DCF_MINUTE ausgefhrt wird)
				}
			}
		}

// --- In der XTAL-Zeit ist eine neue Minute angebrochen
// --- wenn DCF_valid == FALSE, dann ist diese Zeit fr die Anzeige mageblich
// --- ausserdem wird diese Zeit verwendet, wenn keine Synchronisation mit der DCF-Zeit
// --- gefordert ist.

	if (flag & (1<<XTL_MINUTE))
		{
		flag &= (~(1<<XTL_MINUTE));		// das Flag lschen

		if (!(DCF_valid) || !(flag & (1<<SYNCH_ENABLE)))
			{
			DCF_incr_time(&time[CLK][0], 1);
			time[CLK][SEKUNDE] = 0;
			flag |= (1<<DISPLAY);			// Flag veranlasst die Anzeige der Zeit
		#ifdef LOG_RS232
			put_s(" XTL: ");					// Hinweis ausgeben, dass die XTL-Minute Auslser war
		#endif
			}
		}

// --- In der DCF-Zeit wurde der Minutenbeginn erkannt
// --- wenn DCF_valid  == TRUE vorliegen, dann ist diese Zeit mageblich
// --- es sei denn, dass das Flag SYNCH_ENABLE nicht gesetzt ist

	if (flag & (1<<DCF_MINUTE))
		{
		flag &= (~(1<<DCF_MINUTE));		// das Flag lschen
													// wenn kein grundstzlicher Fehler vorliegt
		if ( (DCF_valid) && (flag & (1<<SYNCH_ENABLE)) )	// und das Flag SYNCH_ENABLE gesetzt ist				
			{
													// wenn aber nur ein unvollstndiger Bit-Satz vorliegt, kann kein Fehler
			if(!(DCF_bits_complete()))		// erkannt worden sein, daher hier noch eine Prfung der DCF-Bits.
				{									// DCF_valid bergibt im Fehlerfall die Kontrolle an die xtal-Uhr.
				DCF_valid = FALSE;			// Wenn die Sekunden > 59, dann wurde das Flag XTL_MINUTE bereits gesetzt,
													// wegen DCF_valid hat es aber keine Aktion ausgelst.
													// daher Sekunden prfen und ggf. das Flag noch einmal setzen.
													// zhlen die Sekunden weniger als 60, dann schlgt das Flag via Xtal-Zeit
													// von sich aus zu.
				if (time[CLK][SEKUNDE] > 59) flag |= (1<<XTL_MINUTE);
				}

			else 									// wenn alle DCF-Bits empfangen wurden
				{
				synch_time = SYNCH_TIME;	// wieder auf den Vorgabewert einstellen
													// synch_time muss > 0 sein, wenn das Flag SYNCH_ENABLE gelscht werden soll
													// Das Flag wird erst bei der ANzeige der Minute gelscht, damit nicht 
													// eine eventuell folgendes Flag XTL_MINUTE noch einmal die Zeit incrementiert..
			#ifdef LOG_RS232
				put_s(" DCF: ");				// Hinweis ausgeben, dass die DCF-Minute Auslser war
			#endif
													// Wenn kein Fehler vorlag, xtal-Uhr auf Minutenbeginn
				TCNT0 = 0;						// synchronisieren
				T0_ticks = T0_PRELOAD;		// Timer0 decrementiert T0_ticks auf 0 fr 1 Sekunde
				time[CLK][SEKUNDE] = 0;
				DCF_incr_time(&time[CLK][0], 1);

													// wenn die XTL- und DCF-Zeiten nicht bereinstimmen,
													// hier die DCF-Zeit bernehmen
				if (!(DCF_compare_time(&time[CLK][0], &time[DCF][0])))
					{
					DCF_copy_time(&time[CLK][0], &time[DCF][0]);
					flag |= (1<<SYNCH);		// Nur fr das Debuggen Hinweis auf Neusychronisation setzen
					}								// nicht in error ablegen, sonst wird DCF_valid auf FALSE gesetzt
													// und es wird ein weiterer Minutenbeginn durch die xtal-Zeit mglich
				flag |= (1<<DISPLAY);		// Flag veranlasst die Anzeige der Zeit
				}
			}
		DCF_reset_bits();						// die DCF-bits zurcksetzen/auf den Minutenbeginn synchronisieren
													// darf erst nach dem Aufruf von DCF_bits_complete() erfolgen !
		}

// --- Nach 58 eingegangenen DCF-Bits wird das Signal decodiert und auf Fehler geprft
// --- nur wenn der Vergleich mit den vorausgegangenen 2 Minuten plausibel ist, wird DCF_valid auf TRUE gesetzt
// --- dumm ist es nur, wenn DCF-Bits 'verschluckt' wurden oder zuviele empfangen wurden

	if (flag & (1<<DCF_RX_58))				// das Flag wird in DCF_save_bit() gesetzt, wenn 58 Bit eingegangen sind
		{
		flag &= (~(1<<DCF_RX_58));			// das Flag lschen

		if(!(DCF_bits_complete())) 		// wenn nicht genau 58 Bit empfangen wurden
			{										//
			DCF_valid = FALSE;				// dann bernimmt die xtal-Uhr
			error |= (1<<ERR_BIT_CNT);		// Fehlerflag setzen
			}
		else
			{
			DCF_decode();						// nun steht die decodierte Zeit der neuen Minute in time[DCF]
			}										// in time[CLK] steht noch die aktuelle Minute
													// error kann noch durch Parittsfehler von DCF_decode() gesetzt werden
													// oder durch fehlerhafte Impulslngen whrend der letzten Minute
	#ifdef LOG_RS232
		print_dcfbits();						// die DCF-Bits der folgenden Minute ausgeben
	#endif
													// die aktuelle Zeit mit den letzten beiden DCF-Minuten vergleichen
		if ( (DCF_time_valid()) &&  (!(error)) ) DCF_valid = TRUE; else DCF_valid = FALSE;
		if (!(error)) DCF_backup();		// legt ein Backup der letzten beiden Minuten an

													// eine Ausnahme, damit der Neustart schneller vonstatten geht
													// das Startflag ist nach einem Neustart gesetzt - und nur dann !!
		if ((flag & (1<<START_FLAG)) && (!(error)))
			{
			flag &= ~(1<<START_FLAG);
			DCF_valid = TRUE;					// das DCF-Signal setzt den Minutenanfang
													// und kopiert die DCF-Zeit in die Uhr
			}
		}

// --- Minuten-Anzeige aktualisieren ------------------------------------------

	if (flag & (1<<DISPLAY))
		{

	#ifdef LOG_RS232
		if (DCF_valid)  put_s("d "); else put_s("x "); // Status von DCF_valid ausgeben
		put_s(utoa(flag,msg, 16));
		print_datum(); 						// die Zeitdaten der aktuellen Minute ausgeben
		if (flag  & (1<<SYNCH)) puts_p(PSTR("resynch "));
	#endif

		flag &= (~(1<<SYNCH | 1<<DISPLAY));	// alle Flags lschen
		if (error)								// wenn in der letzten Minute ein Fehler aufgetreten sind ...
			{										// Fehlerflags erst nach der Anzeige lschen !!
			error_abs++;						// zum Debuggen die Fehler mitzhlen

		#ifdef LOG_RS232
			put_s(" # ");
			put_s(itoa(error_abs, msg, 10));
			puts_p(PSTR(" code: 0x"));
			put_s(itoa(error, msg, 16));
			puts_p(PSTR(", err: "));

			if (error & (1<<ERR_NO_DCF))   	puts_p(PSTR("no_DCF, "));
			if (error & (1<<ERR_TWI))      	puts_p(PSTR("time_changed, "));
			if (error & (1<<ERR_SIGNAL)) 		puts_p(PSTR("timeout, "));
			if (error & (1<<ERR_BIT_CNT))		puts_p(PSTR("bit_cnt, "));
			if (error & (1<<ERR_PARITY))  	puts_p(PSTR("parity"));
		#endif
			}

	#ifdef LCD
		display_time(&time[CLK][0]);		// die komplette Anzeige neu aufbauen
	#ifdef SHOW_ERRORS
		lcd_goto_xy(2, 19);					// Anzeigen, ob das DCF-Signal oder der XTAL die Zeit bestimmt
													// dc = DCF, xt = XTAL
		if (DCF_valid) lcd_puts_p(PSTR("dc")); else lcd_puts_p(PSTR("xt"));
	#endif
	#endif

		error = 0;								// Fehlerflags der letzten Minute zurcksetzen
													// aber erst nach der Aktualisierung des Displays
	#ifdef LOG_RS232
		put_c(13);								// Neue Zeile beginnen am Ende der Debugginginfo einer Minute
	#endif
		}
	}												// while(1)
}


// -----------------------------------------------------------------------------
// INT1 - rising edge = Ende eines Impulses
// Es ist die Dauer des Impulses zu bestimmen (0.1 oder 0.2 Sekunden)
// schreibt das empfangene Bit direkt in das Bitarray.

ISR(INT1_vect)
{
register uint8_t x;
x = T1_ticks;

if      (x <  (MS_100 - TOL1)) return;// krzer als 0.1 Sekunden
													// 0.1 Sekunden
else if (x <  (MS_100 + TOL1)) DCF_save_bit(0); // eine 0 speichern

else if (x <  (MS_200 - TOL1)) return;	// krzer als 0.2 Sekunden
													// 0.2 Sekunden
else if (x <  (MS_200 + TOL1)) DCF_save_bit(1); // eine 1 speichern

}

// -----------------------------------------------------------------------------
// INT0 - falling edge = Start des Impulses
// Es wird geprft, wieviel Zeit seit dem letzten Impulsbeginn vergangen ist,
// es knnen 1 Sekunde (normales DCF_Bit) oder 2 Sekunden (Minutenbeginn) sein.
// Wesentliche Aufgabe des Interrupts ist, den 2-Sekunden-Signalabstand zu Beginn
// einer neuen Minute zu erkennen und den Beginn einer neuen Minute an main() zu signalisieren.
// Sind Empfangsfehler im DCF-Signal aufgetreten, dann bernimmt die XTL-Uhr die Aufgabe,
// den Beginn einer neuen Minute zu definieren.

// Hat auch die Aufgabe, beim Programmstart den Beginn der Zhlung zu starten.
ISR(INT0_vect)
{
register uint8_t x;
x = T1_ticks;									// volatile in Register laden

if      (x <  (MS_1000 - TOL2)) return;// krzer als 1 Sekunde ignorieren

else if (x <  (MS_1000 + TOL2));			// Normalfall 1 Sekunde - nur Counter aktualisieren

else if (x <  (MS_2000 - TOL2)) return;// krzer als 2 Sekunden ignorieren

else if (x <  (MS_2000 + TOL2))			// 2 Sekunden
	{
	flag  |= (1<<DCF_MINUTE );				// Flag fr neue DCF-Minute
	}

else error |= (1<< ERR_NO_DCF);			// lnger als 2 Sekunden, damit ungltig

T1_ticks = 0;									// Counter fr Timer1 zurcksetzen
TCNT1H   = 0;
TCNT1L   = 0;
}

// -----------------------------------------------------------------------------
// wird als interne Zeitreferenz fr den Sekundentakt genutzt
// Prescaler = 64, Preload = 256, Counter 225, 3686800 / (64 * 256 * 225) = 1 Sekunde
// Das Flag 1<<XTL_SEC  wird nach genau 1 Sekunde gesetzt und wird in main() ausgewertet.
// Bei fehlerfreiem DCF-Signal signalisiert das 2-Sek-DCF-Signal den Minutenstart.
// Bei fehlerhaftem DCF-Signal signalisiert die 60. Sekunde des Timer0 den Minutenstart.

ISR(TIMER0_OVF_vect)
{

if (--T0_ticks) return;						// abbrechen solange T0_ticks nicht zu 0 decrementiert wurde
T0_ticks = T0_PRELOAD;						// neuen Startwert setzen

time[CLK][SEKUNDE]++;						// Sekunde incrementieren


if (time[CLK][SEKUNDE] < 60)				// solange < 59 wird nur die Sekunde angezeigt
	{												// hier gibt der Timer den Takt an
	flag |= (1<<XTL_SEC);					// Flag fr neue Sekunde setzen
	}

else 												// (time[CLK][SEKUNDE] == 60)
	{
	flag |= (1<<XTL_MINUTE);				// Flag fr neue Minute durch CLK setzen
	}
}

// -----------------------------------------------------------------------------
// fr die Bestimmung der Impulslngen, wird alle 0,01 Sekunden ausgelst
// bei normaler Funktion werden 100/200 ticks ausgelst
// kommt kein Signal, dann luft ticks ber, in diesem Fall wird der Wert
// konstant auf 255 gehalten.
// An diesem Wert kann erkannt werden, dass ein DCF-Signal berfllig ist.
ISR(TIMER1_COMPA_vect)
{
TCNT1H = 0;
TCNT1L = 0;

T1_ticks++;										// incrementieren bis zum berlauf auf 0
if (!(T1_ticks))
	{
	T1_ticks--;									// wenn T1_ticks berluft, dann zurcksetzen auf 255
	error |= (1<< ERR_SIGNAL);				// fehlendes DCF-Signal
	}
}

// ----------------------------------------------------------------------------
// Dekodiert die gelesenen 58 Bits und schreibt die Daten der nchsten Minute
// ins SRAM in das Array 'DCF_time[]'.
// Die zugehrigen Parittsbits werden berprft, Fehler werden in error markiert.
// Die Ankndigungs- und Zusatzbits werden in DCF_time[ZBITS] abgelegt.
// Die Ankndigungsbit stehen im Signal ab Pos. 15 in der Reihenfolge:
// R, A1, Z1, Z2, A2, S  (wobei S immer = 1).
// Sie werden umsortiert (warum eigentlich ?):
// DCF_zbits:(..|..|..|A1|A2|R |Z1|Z2)
// Sommerzeit(..|..|..|A1|A2|R | 1| 0)
// Winterzeit(..|..|..|A1|A2|R | 0| 1)
// A1 Ankndigungsbit fr Sommerzeit
// A2 Ankndigungsbit fr Schaltsekunde
// R  Reserveantenne aktiv
// Die Sommerzeit wird direkt als Bit.7 in time[CLK][DOW] abgelegt.

void DCF_decode(void)
{
													// zuerst die Zusatzbits nach DCF_time[ZBITS] schreiben
time[DCF][ZBITS]  = DCF_get_bit(16);			// A1
time[DCF][ZBITS] *= 2;						// nach links schieben, Platz fr nchstes Bit machen
time[DCF][ZBITS] += DCF_get_bit(19);			// A2
time[DCF][ZBITS] *= 2;						// nach links schieben, Platz fr nchstes Bit machen
time[DCF][ZBITS] += DCF_get_bit(15);			// R
time[DCF][ZBITS] *= 2;						// nach links schieben, Platz fr nchstes Bit machen
time[DCF][ZBITS] += DCF_get_bit(17);			// Z1
time[DCF][ZBITS] *= 2;						// nach links schieben, Platz fr nchstes Bit machen
time[DCF][ZBITS] += DCF_get_bit(18);			// Z2

time[DCF][MINUTE] = calc_time(21, 21 + 7);// Minute auslesen, Paritt prfen
if (calc_parity(21, 21 + 8)) error |= (1<<ERR_PARITY);

time[DCF][STUNDE] = calc_time(29, 29 + 6);// Stunde auslesen, Paritt prfen
if (calc_parity(29, 29 + 7)) error |= (1<<ERR_PARITY);

time[DCF][TAG] = calc_time(36, 36 + 6);// Tag auslesen

time[DCF][DOW] = calc_time(42, 42 + 3);// Day of week auslesen
													// die Nr. des Tag steht nun als Ziffer in DOW
													// z.B. Mittwoch = 3
													// umwandeln in Bitposition 0 ... 6
													// dazu die 1 um (3-1) mal nach links shiften
time[DCF][DOW] = 1 << (time[DCF][DOW] - 1);

													// die Sommerzeit als Bit.7 in den DOW schmuggeln
if ((time[DCF][ZBITS] & 0x03) == 2) set_summertime(&time[DCF][0], TRUE);

time[DCF][MONAT] = calc_time(45, 45 + 5);// Monat auslesen

time[DCF][JAHR] = calc_time(50, 50 + 8);// Jahr auslesen, Paritt prfen
if (calc_parity(36, 36 + 23)) error |= (1<<ERR_PARITY);
}

// -----------------------------------------------------------------------------
// Decodiert eine Zahl aus einem Bit-Bereich des DCF_Bit_Arrays,
// das ergibt dann den Wert fr entweder: Minute, Stunde, Tag, Monat, Jahr
// Parameter: Startposition des ersten Bits
//		        Endposition   des letzten Bits
// Rckgabe:  der ermittelte Wert
uint8_t calc_time(uint8_t start, uint8_t end)
{
// uint8_t summand[] PROGMEM = { 1, 2, 4, 8, 10, 20, 40, 80 };
// um PROGMEM zu nutzen, darf das Array nicht lokal sein

uint8_t r = 0;
uint8_t i = 0;

while (start < end)
	{
	if (DCF_get_bit(start++))
		{
		r += pgm_read_byte(&summand[i]);
		}
	i++;
	}

return(r);
}

// -----------------------------------------------------------------------------
// Berechnet die Paritt, beginnend an Position "start" ber "cnt" Bits
// (Achtung:  das letzte Bit im Bereich ist das Parittsbit !)
// Parameter: Startpunkt ab Arraybeginn
//		        Letzes Bit = Parittsbit
// Rckgabe:  (0 = kein Parittsfehler, 1 = Parittsfehler)
// Das Parittsbit ergnzt die Summe der einzelnen Bits auf eine gerade Zahl,
// d.h. Bit.0 ist nicht gesetzt - sonst liegt ein Fehler vor.
uint8_t calc_parity(uint8_t start, uint8_t end)
{
register uint8_t p = 0;
while (start < end)
	{
	if (DCF_get_bit(start++)) p++;
	}

return(p & 0x01);
}

// -----------------------------------------------------------------------------
// Ein Zeitarray wird um x Minuten incrementiert.
// ES wird die Zeitzhlung aktualisiert und incl. aller mglichen berlufe.
// Weiterhin erfolgt Prfung auf Wechsel zwischen Sommer-/Winterzeit.
void DCF_incr_time(uint8_t *zeit, uint8_t x)
{
register uint8_t t;

do	{
	zeit[MINUTE]++;
	if (zeit[MINUTE] < 60) continue;		// solange keine volle Stunde angebrochen ist,
													// kann hier beendet werden
	zeit[MINUTE] = 0;
	zeit[STUNDE]++;
													// zu jeder vollen Stunde testen, ob eine Wechsel
	switch_summertime(zeit);				// zwischen Sommer- und Winterzeit ansteht.
													// im jeweiligen Zeitarray wird dann das Flag fr
													// die Sommerzeit (DOW.7) gegebenfalls gendert
													// wenn Stunde == 24, dann kann kein Fehler entstehen,
													// da nur um 2 bzw. 3 Uhr umgeschaltet wird
	if (zeit[STUNDE] < 24) continue;
	zeit[STUNDE] = 0;
	zeit[TAG]++;
													// den Wochentag incrementieren
	t = zeit[DOW] & 0x80;					// Bit.7 im Wochentag markiert die Sommerzeit, zwischenspeichern !
	zeit[DOW] &= 0x7F;						// Bit.7 entfernen
	zeit[DOW] *=2;								// Wochentag incrementieren (nach links schieben)
													// berlauf von Sonntag auf Montag
	if (zeit[DOW] == 0x80) zeit[DOW] = 0x01;
	zeit[DOW] += t;							// Sommerzeit als Bit.7 wieder addieren

													// nun kommen die Tage des Monats in Spiel ...
	if (zeit[MONAT] == 2)					// zuerst Sonderfall Februar und Schaltjahr
		{
		if (is_leap(zeit[JAHR]))
			{
			if (zeit[TAG] <= 29) continue;
			}
		else
			{
			if (zeit[TAG] <= 28) continue;
			}
		}

	else											// die brigen Monate
		{
		if (zeit[TAG] <= pgm_read_byte(&days_of_month[zeit[MONAT]])) continue;
		}

	zeit[TAG] = 1;								// der letzte Tag des Vormonats war erreicht
	zeit[MONAT]++;
	if (zeit[MONAT] > 12)
		{
		zeit[MONAT] = 1;
		zeit[JAHR]++;
		}
	} while(--x);								// den Rundgang fr jede Minute wiederholen
}

// ----------------------------------------------------------------------------
// Die fr die Bestimmung des Sonnenstandes bentigte UTC-Zeit liegt in unserer
// Zeitzone im Winter um 1 Stunde, whrend der Sommerzeit um 2 Stunden vor unserer Zeit
// Kopiert die CLK_time auf die UTC_time und zhlt je nach cet/cest um eine bzw. zwei
// Stunden zurck.
void DCF_make_UTC_time(void)
{
DCF_copy_time(&time[UTC][0], &time[CLK][0]);	// die Zeit kopieren

DCF_decr_time(&time[UTC][0], 60);				// um 60 Minuten decrementieren
if (is_summertime(&time[CLK][0]))				// und whrend der Sommerzeit um eine weitere Stunde
	{
	DCF_decr_time(&time[UTC][0], 60);			// um noch einmal 60 Minuten decrementieren
	}
}

// ----------------------------------------------------------------------------
// decrementiert ein Zeitarray um x Minuten
// *zeit ist das Zeitarray, das bearbeitet werden soll
void DCF_decr_time(uint8_t *zeit, uint8_t x)
{
register uint8_t t;

do	{
	if (zeit[MINUTE]) 							// solange die Minute > 0 gibt es keinen berlauf
		{
		zeit[MINUTE]--;
		continue;									// das war's schon
		}
	else
		{
		zeit[MINUTE] = 59;
		zeit[STUNDE]--;
		}

	if (zeit[STUNDE] == 255) 					// auf berlauf testen
		{
		zeit[STUNDE] = 23;
		}
	else
		{
		continue;
		}

	// den Wochentag decrementieren
	t = zeit[DOW] & 0x80;						// das Bit fr die Sommerzeit sichern
	zeit[DOW] &= 0x4F;							// die Sommerzeit entfernen
	zeit[DOW] /= 2;								// den DOW decrementieren
														// auf underflow testen
	if (!(zeit[DOW])) zeit[DOW] = 0x40; 	// wenn == 0, dann DOW.6 setzen
	zeit[DOW] += t;								// Sommerzeit wieder addieren

	// den Test auf eine nderung der Sommerzeit erspare ich mir an dieser Stelle ...
	// ist fr den Sonnenstand nicht von Bedeutung, da die UTC-Zeit massgeblich ist
	if (zeit[TAG] > 1)
		{
		zeit[TAG]--;
		continue;
		}

	else												// bei Underflow des Tages zuerst den
		{												// neuen Monat bestimmen
		zeit[MONAT]--;
		if (zeit[MONAT] == 0)
			{
			zeit[MONAT] = 12;
			zeit[JAHR]--;
			}											// nun haben wir die Anzahl der Tage
		zeit[TAG] = pgm_read_byte(days_of_month[zeit[MONAT]]);
														// falls aber Februar und Schaltjahr
		if ((zeit[MONAT] == 2) && (is_leap(zeit[JAHR]))) zeit[TAG]++;
		}
	} while (--x);
}

// -----------------------------------------------------------------------------
// kopiert ein Zeit-Array
void DCF_copy_time(uint8_t *target, uint8_t *source)
{
register uint8_t i = 6;							// Element mit Index 5 ... 0 kopieren
do	{
	i--;
	target[i] = source[i];
	} while(i);
}

// -----------------------------------------------------------------------------
// vergleicht die Bytes der Arrays time1 und time2 byteweise miteinander
// liefert: TRUE  wenn identisch,
//          False wenn ungleich
uint8_t DCF_compare_time(uint8_t *time1, uint8_t *time2)
{
uint8_t i = 6;										// Element mit Index 5 ... 0 vergleichen
uint8_t r = TRUE;
do {
	i--;												// ein einziger fehlerhafter Vergleich reicht aus
	if (! (time1[i] == time2[i]) ) r = FALSE;
	} while (i);
return(r);
}

// -----------------------------------------------------------------------------
// Wird sinnvollerweise jeweils zu einer vollen Stunde aufgerufen, da nur dann ein Wechsel
// zwischen Sommer- und Winterzeit stattfinden kann.
//
// Prfen anhand von Datum und Zeit der internen Uhr, ob zwischen Sommer- und Winterzeit umgeschaltet werden muss.
//
// Am letzten Sonntag im Mrz um 02.00h wird die Uhr auf 03.00h,
// am letzten Sonntag im Oktober um 03.00h wird die Uhr auf 02.00h gestellt.
// Ausgehend von der aktuellen Sommerzeit wird geprft,
// wenn Sommerzeit = 0, ob Monat  = 3, Wochentag = 7 und Datum zwischen 25 und 31 und Stunde = 2
// wenn Sommerzeit = 1, ob Monat = 10, Wochentag = 7 und Datum zwischen 25 und 31 und Stunde = 3
// zur Kontrolle/Besttigung knnte noch das Ankndigungsbit A1 (Bit 16) herangezogen werden ...
// kritisch ist Umschaltung im Oktober, da hier der Zeitraum von 0.200h bis 03.00h zweimal nacheinander
// durchlaufen wird -> hier die aktuelle Einstellung der Sommerzeit mageblich:
// wenn Sommerzeit, dann Umschaltung, wenn (schon) Winterzeit, dann keine Umschaltung !
void switch_summertime(uint8_t *zeit)
{
if (zeit[TAG] > 24)							// Gewechselt wird zwischen dem 25. und dem 31. (Mrz bzw. Oktober)
	{
	if ((zeit[DOW] & 0x40) )				// nur an einem Sonntag (dann ist Bit.6 gesetzt)
		{
		if (zeit[MONAT] == 3)				// im Mrz
			{
			if (zeit[STUNDE] == 2)			// um 2:00h die Uhr auf 3:00h vorstellen
				{
				zeit[STUNDE] = 3;				// Zeit vorstellen auf 3;00h
				set_summertime(zeit, TRUE);// auf Sommerzeit schalten
				}
			}
													// im Oktober die Uhr zurckstellen (sofern noch Sommerzeit !)
													// 03.00h wird zweimal durchlaufen,
													// beim 2. Mal ist aber bereits auf Winterzeit umgeschaltet,
													// daher wird die Uhr nicht noch einmal zurckgestellt
		else if ((zeit[MONAT] == 10) && (is_summertime(zeit)))
			{
			if (zeit[STUNDE] == 3)
				{
				zeit[STUNDE] = 2;				// Zeit zurckstellen auf 2:00h
				set_summertime(zeit, FALSE);// und auf Winterzeit stellen
				}
			}

		}
	}
}

// ----------------------------------------------------------------------------
// Prft, ob das DCF-Signal Sommerzeit signalisiert
// Rckgabe: 1 = Sommerzeit
//           0 = Keine Sommerzeit
// es wird nicht DCF_zbits ausgewertet, sondern das Bit.7 in time[DOW] !
uint8_t is_summertime(uint8_t *zeit)
{
return(zeit[DOW] & 0x80); 					// Bit.7 in DOW signalisiert die Sommerzeit
}

// -----------------------------------------------------------------------------
// setzt/lscht das Sommerzeitflag in time[DOW].7
void set_summertime(uint8_t *zeit, uint8_t i)
{
if (i) zeit[DOW] |= 0x80; else zeit[DOW] &= ~(0x80);
}

// -----------------------------------------------------------------------------
// ein Schaltjahr liegt vor, wenn die Jahreszahl durch 4, 100 (oder 400) ohne Rest teilbar
// es werden nur die beiden Flle mod 4/ mod 100 untersucht

uint8_t is_leap(uint8_t j)
{													// Es liegt ein Schaltjahr vor,
if (!(j & 0x03)) return(TRUE);			// wenn mod 4 keinen Rest ergibt

while (j > 100) j -= 100;					// diese berprfung ist hier wohl eher berflssig ..
if (!(j)) return(TRUE);						// wenn mod 100 keinen Rest ergibt

return(FALSE);
}


// -----------------------------------------------------------------------------
// Ermittelt die lfd. Nummer des aktuellen Tages im Jahr unter Bercksichtigung von Schaltjahren
uint16_t day_of_year(uint8_t *zeit)
{
uint16_t z = 0;
uint8_t i = 1;

while (i < zeit[MONAT])						// die vollen Tage der Monate addieren
	{
	z+= pgm_read_byte(&days_of_month[i++]);
	if (i == 2)									// im Februar das Schaltjahr bercksichtigen
		{
		if (is_leap(zeit[JAHR])) z++;		// und einen Tag addieren fr den 29. Februar
		}
	}
z += zeit[TAG];								// nun noch die Tage des laufenden Monats addieren
return(z);
}

// -----------------------------------------------------------------------------
// bei Anbruch einer neuen Sekunde nur die Sekundenanzeige aktualisieren
#ifdef LCD
void display_second(void)
{
lcd_goto_xy(1,19);
lcd_dezprint(time[CLK][SEKUNDE]);		// Sekunde dezimal ausgeben
}
#endif

// -----------------------------------------------------------------------------
// die gesamte Zeitanzeige Stunden, Minuten (und Sekunden) aktualisieren
#ifdef LCD
void display_time(uint8_t *zeit)
{

register uint8_t i = 0;
register uint8_t j;

lcd_goto_xy(1, 1);

j = (zeit[DOW] & 0x7F);						// Bit.7 ignorieren -> Flag fr Sommerzeit
													// der Tag ist als Bit 1 ..6 codiert !

while (j)										// Index des Textstrings bestimmen
	{												// dazu die binrzahl solange nach rechts
	j = j >> 1;
	if (j) i++;
	}
i = i * 2;										// es werden 2 Char je Tag benutzt

lcd_putc(pgm_read_byte(&day_names[i++]));
lcd_putc(pgm_read_byte(&day_names[i]));
lcd_putc(' ');

lcd_dezprint(zeit[TAG]);
lcd_putc('.');
lcd_dezprint(zeit[MONAT]);
lcd_putc('.');
lcd_dezprint(zeit[JAHR]);

lcd_putc(' ');

lcd_dezprint(zeit[STUNDE]);
lcd_putc(':');
lcd_dezprint(zeit[MINUTE]);
lcd_putc(':');
lcd_dezprint(time[CLK][SEKUNDE]);

lcd_goto_xy(2,6);							// den Unterstrich in Zeile 2 ziehen
i = 15;										// nicht in dem Bereich, in dem die Alarmdefs
while (i--)									// angezeigt werden (Spalte 1-5)
	{
	lcd_putc('_');
	}

lcd_goto_xy(2,13);						// Position fr Fehler/Sommmer/Winterzeit
if (error) 									// wenn in der letzten Minute ein Fehler auftrag
	{
	lcd_puts_p(PSTR("er:"));			// Fehlermeldung
	lcd_hexprint(error);					// Fehlercode hexadezimal
	}

else											// ohne Fehler wird Hinweis auf Sommerzeit ausgegeben
	{
	if (is_summertime(&time[CLK][0])) lcd_puts_p(PSTR("cest")); else lcd_puts_p(PSTR("cet"));
	}

#ifdef SHOW_ERRORS
char space[] = "  ";
lcd_goto_xy(3, 1);
if (time[DCF][ZBITS] & ~(0x03))		// wenn Ankndigungsbits gesendet wurden
												// die Bits fr Sommer-/Winterzeit ignorieren
	{
	if (!(error))							// nur Anzeigen, wenn das Datum fehlerfrei
		{
		lcd_hexprint(time[DCF][ZBITS]);
		}
	}

else
	{
	lcd_puts(space);						// den Bereich der Meldung lschen
	}

if (error_abs)								// absolute Anzahl der Fehler ausgeben
	{
	lcd_goto_xy(4, 1);

	if ((time[CLK][MINUTE] == 0) && (time[CLK][STUNDE] == 0)) // Mitternacht
		{
		lcd_puts(space);					// um Mitternacht den Zhler und die
		error_abs = 0;						// Anzeige der Fehler lschen

		lcd_goto_xy(3, 1);				// auch die Anzeige der Ankndigungsbits lschen
		lcd_puts(space);
		}

	else
		{
		lcd_hexprint(error_abs);		// Anzahl der kumulierten Fehler des Tages ausgeben
		}
	}

#endif

}
#endif


// ----------------------------------------------------------------------------
// Vergleicht mit den letzten beiden MInuten
uint8_t DCF_time_valid(void)
{
return( DCF_compare_time(&time[BK2][0], &time[BK1][0]) && DCF_compare_time(&time[BK1][0], &time[DCF][0]) );
}


// ----------------------------------------------------------------------------
// legt eine Kopie der letzten beiden Minuten ab und incrementiert die Zeit so,
// dass ein direkter Abgleich mit der neuen Minute mglich ist
// Zuerst Vergleichen, dann Sichern
void DCF_backup(void)
{
DCF_copy_time(&time[BK2][0], &time[BK1][0]);
DCF_copy_time(&time[BK1][0], &time[DCF][0]);

DCF_incr_time(&time[BK2][0],1);
DCF_incr_time(&time[BK1][0],1);
}

#ifdef LOG_RS232
// ----------------------------------------------------------------------------
// gibt zum Debuggen die DCF-Bits ab Bit.15 der nchsten Minute via RS232 aus
// als 0/1, die logischen Gruppen durch einen '.' getrennt
// positive Parittsprfung wird mit '!' dargestellt,
// Fehler in der Parittsprfung werden mit '?' markiert
void print_dcfbits(void)
{
uint8_t i, z;

// Ankndigungsbits -----------------------------------------------------------
for (i = 15; i < 21; i++)
{
	if (DCF_get_bit(i))
	{
		put_c('1');
		z++;
	}
	else put_c('0');
}
// Paritt

// Trenner
put_c('.');

// Minuten --------------------------------------------------------------------
z = 0;
for (i = 21; i < 28; i++)
{
	if (DCF_get_bit(i))
	{
		put_c('1');
		z++;
	}
	else put_c('0');
}
// Paritt
z += DCF_get_bit(28);
if (z & 1) put_c('?'); else put_c('!');	// Paritt muss immer gerade Zahl sein
// Trenner
put_c('.');

// Stunden --------------------------------------------------------------------
z = 0;
for (i = 29; i < 35; i++)
{
	if (DCF_get_bit(i))
	{
		put_c('1');
		z++;
	}
	else put_c('0');
}
// Paritt
z += DCF_get_bit(35);
if (z & 1) put_c('?'); else put_c('!');
// Trenner
put_c('.');

// der Rest --------------------------------------------------------------------
z = 0;
for (i = 36; i < 58; i++)
{
	if (i == 42) put_c('.');
	if (i == 45) put_c('.');
	if (i == 50) put_c('.');
	if (DCF_get_bit(i))
	{
		put_c('1');
		z++;
	}
	else put_c('0');
}
// Paritt
z += DCF_get_bit(58);
if (z & 1) put_c('?'); else put_c('!');

// Ankndigungsbits im Klartext -----------------------------------------------

if (time[DCF][ZBITS] & ~(0x03))				// wenn Ankndigungsbits im DCF-Signal gesendet wurden
														// die Bits fr Sommer-/Winterzeit ignorieren
	{
	if (!(error))									// nur Anzeigen, wenn das Datum fehlerfrei
		{
		if (time[DCF][ZBITS] & 0x10)	puts_p(PSTR(", A1"));
		if (time[DCF][ZBITS] & 0x08)	puts_p(PSTR(", A2"));
		if (time[DCF][ZBITS] & 0x04)	puts_p(PSTR(", R"));
		}
	}
}

// ----------------------------------------------------------------------------
// gibt zum Debuggen die Zeit der aktuellen Minute via RS232 aus
void print_datum(void)
{
put_c(' ');

if (time[CLK][STUNDE] < 10) put_c('0');	// Stunde und  Minuten
put_s(utoa(time[CLK][STUNDE], msg , 10));
put_c(':');
if (time[CLK][MINUTE] < 10) put_c('0');
put_s(utoa(time[CLK][MINUTE], msg, 10));

put_c('-');											// Sommer-/Winterzeit
if (is_summertime(&time[CLK][0])) put_c('s'); else put_c('w');
put_c('-');

if (time[CLK][TAG] < 10) put_c('0');		// Tag, Monat, Jahr
put_s(utoa(time[CLK][TAG],msg, 10 ));
put_c('.');
if (time[CLK][MONAT] < 10) put_c('0');
put_s(utoa(time[CLK][MONAT], msg, 10));
put_c('.');
if (time[CLK][JAHR] < 10) put_c('0');
put_s(utoa(time[CLK][JAHR], msg, 10));

put_c(' ');
}
#endif
// --- [ eof ] ----------------------------------------------------------------
